import 'package:meta/meta.dart';

import '../blocks/blocks.dart';
import '../errors.dart';
import 'adapter.dart';
import 'registry.dart';

class RegistryError extends TapeError {
  RegistryError(List<ErrorPart> parts) : super(parts);
}

class AdapterError extends TapeError {
  AdapterError(List<ErrorPart> parts) : super(parts);
}

class TaperRegisteredForMultipleTypeCodes extends RegistryError {
  TaperRegisteredForMultipleTypeCodes(List<ErrorPart> parts) : super(parts);
}

class TaperAlreadyRegisteredError extends RegistryError {
  TaperAlreadyRegisteredError({
    @required Registry registry,
    @required this.adapter,
    @required this.id,
  }) : super([
          ErrorSummary('$adapter already registered for id $id.'),
          ErrorText('You tried to register $adapter for id $id, but it was '
              "already registered under that id. Here's a tree of all "
              'currently registered adapters:\n'
              '${registry.treeAsDebugString()}'),
          ErrorCallToAction("Don't register $adapter twice."),
        ]);

  final TapeAdapter<dynamic> adapter;
  final int id;
}

class AdapterAlreadyRegisteredForDifferentIdError extends RegistryError {
  AdapterAlreadyRegisteredForDifferentIdError({
    @required Registry registry,
    @required this.adapter,
    @required this.oldId,
    @required this.newId,
  }) : super([
          ErrorSummary('$adapter already registered for id $oldId.'),
          ErrorText('You tried to register $adapter for id $newId, but it '
              "was already registered under id $oldId. Here's a tree of all "
              'currently registered adapters:\n'
              '${registry.treeAsDebugString()}'),
          ErrorCallToAction("Don't register $adapter twice."),
        ]);

  final TapeAdapter<dynamic> adapter;
  final int oldId;
  final int newId;
}

class IdAlreadyInUseError extends RegistryError {
  IdAlreadyInUseError({
    @required Registry registry,
    @required this.adapter,
    @required this.id,
    @required this.adapterForId,
  }) : super([
          ErrorSummary('Id $id is already in use.'),
          ErrorText('You tried to register $adapter for id $id, but '
              "$adapterForId is already registered for that id. Here's a "
              'tree of all currently registered adapters:\n'
              '${registry.treeAsDebugString()}'),
          ErrorCallToAction("Don't register $adapter and $adapterForId for "
              'the same id.'),
        ]);

  final TapeAdapter<dynamic> adapter;
  final int id;
  final TapeAdapter<dynamic> adapterForId;
}

class UnexpectedBlockError extends AdapterError {
  UnexpectedBlockError(this.block, this.expectedType)
      : super([
          ErrorSummary(
              'Expected $expectedType, but found ${block.runtimeType}.'),
          ErrorText("Here's the block that was found:\n$block"),
          ErrorCallToAction("If you're writing an adapter, you probably have "
              'an error in it.'),
          ErrorCallToAction('Otherwise, consider opening an issue: '
              '$repositoryLink/issues/new?/issues/new?template=2-bug.md&'
              'labels=bug&title=UnexpectedBlockError+thrown'),
        ]);

  final Block block;
  final Type expectedType;

  @override
  String toString() =>
      'Expected $expectedType, but found ${block.runtimeType}:\n$block';
}
